install.packages("synapser", repos=c("http://ran.synapse.org", "http://cran.fhcrc.org"))
install.packages("dplyr", repos = "http://cran.r-project.org")
install.packages("argparse", repos = "http://cran.r-project.org")
install.packages("rmarkdown", repos = "http://cran.r-project.org")
install.packages("UpSetR", repos = "http://cran.r-project.org")
install.packages("testthat", repos = "http://cran.r-project.org")
install.packages("xtable", repos = "http://cran.r-project.org")
# ggpubr is used in the data guide generation
install.packages("ggpubr", repos = "http://cran.r-project.org")
install.packages("XML", repos = "http://cran.r-project.org")
install.packages("BiocManager", repos = "http://cran.r-project.org")

BiocManager::install("VariantAnnotation")

library(synapser)
library(dplyr)
library(argparse)
library(UpSetR)
library(rmarkdown)
library(testthat)
library(VariantAnnotation)
library(xtable)
library(ggpubr)
